/**
  Copyright 2002 Bruce Spencer 
*/
import java.util.*;
import java.io.*;
import org.jdom.JDOMException;

/**
   Discrimination trees, and the access iterator associated with
   them, provide a means to quickly refer to a list of items,
   each of which is associated with a particular first order term
   structure, called it accessor term.  Given a term structure
   used as the query part of the access, the "target term" if you
   like, there are three iterators that could be used to access
   the items: LessGeneralIterator, MoreGeneralIterator and
   UnifiableIterator.  These return iterators that give the items
   associated with accessor terms that are, respectively, less
   general, more general and unifiable with the target term.  The
   indexing data structure used is the discrimination tree.  It
   is constructed (as if) by listing all of the accessor terms
   and combining like prefixes, to construct a tree.  In this
   implementation we refer to DiscTrees, really just a specific
   type of discrimination tree, built for our purposes.

   A DiscTree is a rooted directed tree of DiscNodes where each
   DiscNode refers to its k children DiscNodes as child 0, ...,
   child k-1.  The number k may vary from DiscNode to DiscNode.
   A DiscNode which is a leaf node, here called a DiscLeafNode,
   refers to a list of items, which are the things we are trying
   to find.  Insertion of a pair consisting of an item and a
   sequence of numbers (i_1, ..., i_n) where 0<=i<k for each i,
   consists of, for each j=1,...,n, following the i_j'th child
   DiscNode, or if no such child exists, creating it, and then
   placing the item at the resulting leaf DiscNode.  No sequence
   in such an insertion pair is a proper prefix of any other such
   sequence.  Thus arities of function symbols are strictly
   observed.  Removal of an item/sequence pair consists of
   finding the leaf as in the insertion step, and removing the
   item from that list in the leaf, if it exists. No removals
   from internal nodes is done.

   If an insertion is attempted to a child j and the array of
   children for this particular DiscNode has k<j children, then
   the array of children is increased to the current size of the
   symbolTable, the old array is copied into the new array, the
   old array is deleted and the references are diverted to the
   new array.




  Created: Mon Nov 13 19:51:13 2000
 
  @author Bruce Spencer

 */
public class DiscTree {
  
  /**
     The variable <code>root</code> stores the internal node of the DiscTree.
  */
  private DiscInteriorNode root;

  public DiscTree(){
    symbolTable = new SymbolTable();
    root = null;
  }

  public SymbolTable symbolTable;

  public void insert(DefiniteClause clause){
    if(root == null){
      root = new DiscInteriorNode();
    }
    root.insert(clause);
  }

  public Vector retrieve(DCTree.Goal goal) throws NoSuchDiscNodeException{
    if(root == null)
      throw new NoSuchDiscNodeException();
    return root.retrieve(goal);
  }

  public UnifiableIterator unifiableIterator(DCTree.Goal goal){
    return new UnifiableIterator(goal);
  }

  public String toString(){
    if(root == null)
      return null;
    return "root"+root.toString(0);
  }
  
  public void parseDCFile(String fileName) throws IOException, ParseException {
    DCFileParser dcfp = new DCFileParser(symbolTable);
    dcfp.parseDCFile(fileName);
    Iterator cit = dcfp.iterator();
    while(cit.hasNext()){
      //      insert((DefiniteClause) cit.next());
      DefiniteClause clause = (DefiniteClause) cit.next();
      if(DCTree.INTERNAL_TRACE)
	System.out.println("Inserting clause " + clause);
      insert(clause);
    }
  }

  public void parseRuleMLFile(String fileName) throws IOException, JDOMException {
    RuleMLFileParser rmlfp = new RuleMLFileParser(symbolTable);
    rmlfp.parseRuleMLFile(fileName);
    Iterator cit = rmlfp.iterator();
    while(cit.hasNext()){
      //insert((DefiniteClause) cit.next());
      DefiniteClause clause = (DefiniteClause) cit.next();
      System.out.println("Inserting clause " + clause);
      insert(clause);
    }
  }

  public DefiniteClause parseQueryString(String queryString) throws IOException, ParseException{
    DCFileParser dcfp = new DCFileParser(symbolTable);
    DefiniteClause clause = dcfp.parseQueryClause(queryString);
    //symbolTable.dump();
    return clause;
  }

  public int symbolTablePosition(String symbol, int arity){
    return symbolTable.internSymbol(symbol, arity);
  }

  private static String indentString(int indent){
    StringBuffer s = new StringBuffer("\n");
    for(int i = 0; i < indent; i++){
      s.append("  ");
    }
    return new String(s);
  }

  /**
     The interface <code>DiscNode</code> describes the three things we
     might want to do with DiscNodes.  
  */
  interface DiscNode {
    void insert(DefiniteClause clause);
    Vector retrieve(DCTree.Goal goal) throws NoSuchDiscNodeException;
    public String toString(int indent);
  }

  class DiscLeafNode implements DiscNode{
    Vector items;
    DiscLeafNode(){
      items = new Vector();
    }
    public void insert(DefiniteClause clause){
      items.add(clause);
    }
    public Vector retrieve(DCTree.Goal goal) 
      throws NoSuchDiscNodeException{
      return items;
    }
    public String toString () {
      return toString(0);
    }
    public String toString (int indent) {
      StringBuffer s = new StringBuffer();
      s.append(DiscTree.indentString(indent));
      s.append("<");
      if(items.size()>0){
	for(int i = 0; i<items.size()-1; i++){
	  s.append(items.get(i));
	  s.append(",");
	}
	s.append(items.get(items.size()-1));
      }
      s.append(">");
      return s.toString();
    }
  }

  public class DiscInteriorNode implements DiscNode{

    DiscNode[] child;
    
    DiscInteriorNode(){
      child = new DiscNode[symbolTable.size()];
    }


    public void insert(DefiniteClause clause){
      //assumes symbols has at least one element
      insert(clause.symbols, 0, clause.symbols[0][1], clause);
    }

    public void insert(int[][] symbols, int from, int to, DefiniteClause clause){
      int at = symbols[from][0];
      if(at<0){
	//this is a variable.  Since these are imperfect discrimination trees
	// we map all variables to 0
	at = 0;
      }
      if(at>=child.length){
	//The array of children is too small, because the
	//symbolTable has been increased in the mean time.  We
	//assume that value of at should will never exceed the
	//position of the last symbol in the symbolTable.
	DiscNode[] newChild = new DiscNode[symbolTable.size()];
	System.arraycopy(child, 0, newChild, 0, child.length);
	child = newChild;
      }
      if(child[at] == null){
	if(from+1 == to)
	  child[at] = new DiscLeafNode();
	else
	  child[at] = new DiscInteriorNode();
      }
      if(from+1 == to)
	((DiscLeafNode) child[at]).insert(clause);
      else
	((DiscInteriorNode) child[at]).insert(symbols, from+1, to, clause);
    }

    public Vector retrieve(DCTree.Goal goal) throws NoSuchDiscNodeException{
      //assumes goal has at least one element
      return retrieve(goal, goal.symbolIndex, 
		      goal.symbolIndex+goal.goalList.symbols[goal.symbolIndex][1]);
    }


    public Vector retrieve(DCTree.Goal goal, int from, int to) 
      throws NoSuchDiscNodeException{
      int at = goal.goalList.symbols[from][0];
      if(at<0){
	//this is a variable.  Since these are imperfect discrimination trees
	// we map all variables to 0
	at = 0;
      }
      if(child[at] == null)
	throw new NoSuchDiscNodeException();
      if(from+1 == to)
	return ((DiscLeafNode) child[at]).retrieve(null); 
      else
	return ((DiscInteriorNode) child[at]).retrieve(goal, from+1, to);
    }

    public String toString(){
      return toString(0);
    }

    public String toString(int indent){
      StringBuffer b = new StringBuffer();
      b.append(DiscTree.indentString(indent));
      for(int i = 0; i < child.length; i++){
	if(child[i]!=null){
	  b.append(symbolTable.printName(i));
	  b.append("{");
	  b.append(child[i].toString(indent+1));
	  b.append(DiscTree.indentString(indent));
	  b.append("}");
	}
      }
      return b.toString();
    }
  }
  
  class UnifiableIterator implements Iterator{

    UnifiableIterator (DCTree.Goal goal){
      n = root;
      symbols = goal.goalList.symbols;
      from = goal.symbolIndex;
      to = from+symbols[from][1];
      s = from;
      if(n==null)
	s = to;//this is a simple method (but a kludge) to deal with empty trees
    }

    private DiscTree.DiscNode n;
    private int[][] symbols;
    private int from;
    private int to;
    private Stack back = new Stack();
    private int s;
    private boolean recentRestore = false;
    private boolean foundNext = false;
    private Iterator leafIterator = null;
    private EndTermIterator e = null;

    public boolean hasNext(){
      if(foundNext)
	return true;
      else if (leafIterator != null && leafIterator.hasNext()){
	foundNext = true;
	return true;
      }
      else {
	foundNext = false;
	leafIterator = null;
	if( s==to){
	  if (back.empty() ) {
	    return false;
	  }
	  else {
	    if(back.peek() instanceof StackFrameLG){
	      StackFrameLG restore = (StackFrameLG) back.pop();
	      s = restore.s;
	      e = restore.e;
	      n = (DiscTree.DiscNode) e.next();
	      if(e.hasNext()) 
		back.push(new StackFrameLG(s, e));
	      e = null;
	    }
	    else {
	      StackFrameMG restore = (StackFrameMG) back.pop();
	      recentRestore = true;
	      s = restore.s;
	      n = restore.n;
	    }
	  }
	}
	DiscTree.DiscNode matchingChild;
	while(s<to){
	  int sChildIndex= symbols[s][0];
	  if(sChildIndex<0) 
	    sChildIndex = 0;
	  if(!recentRestore && ((DiscTree.DiscInteriorNode)n).child[0]!=null 
	     && sChildIndex != 0){
	    StackFrameMG pair = new StackFrameMG(s, n);
	    back.push(pair);
	    n = ((DiscTree.DiscInteriorNode)n).child[0];
	    s += symbols[s][1];
	  }
	  else{ 
	    recentRestore = false;
	    //recall that sChildIndex= symbols[s][0];
	    if(sChildIndex > ((DiscTree.DiscInteriorNode)n).child.length)
	      matchingChild = null;
	    else
	      matchingChild = ((DiscTree.DiscInteriorNode)n).child[sChildIndex];
	    if((matchingChild != null && sChildIndex != 0)){
	      recentRestore = false;
	      n = matchingChild;
	      s++;
	    }
	    else if(sChildIndex == 0 && e == null){
  	      //EndTermIterator eNew = new EndTermIterator(1,n);
  	      //while(eNew.hasNext()){
  	      //  System.out.println("New end term iterator " + eNew.next());
  	      //}
	      e = new EndTermIterator(1,n);
	      if (e.hasNext()){
		n = (DiscTree.DiscNode) e.next();
		s++;
		if(e.hasNext())
		  back.push(new StackFrameLG(s, e));
	      }
	      e = null;
	    }
	    else if(!back.empty()){
	      if((back.peek()) instanceof StackFrameLG){
		StackFrameLG restore = (StackFrameLG) back.pop();
		s = restore.s;
		e = restore.e;
		n = (DiscTree.DiscNode) e.next();
		if(e.hasNext()) 
		  back.push(new StackFrameLG(s, e));
		e = null;
	      }
	      else {
		StackFrameMG restore = (StackFrameMG) back.pop();
		recentRestore = true;
		s = restore.s;
		n = restore.n;
	      }
	    }
	    else{
	      foundNext = false;
	      return false;
	    }
	  }
	}
	try{
	  leafIterator = n.retrieve(null).iterator();
	  foundNext = true;
	  return true;
	}
	catch (DiscTree.NoSuchDiscNodeException e){
	  System.err.println("This should not happen.");
	  e.printStackTrace();
	  return false;
	}
      }
    }
  
    public Object next() throws NoSuchElementException {
      if(!hasNext())
	throw new NoSuchElementException();
      foundNext = false;
      return leafIterator.next();
    }

    public void remove() throws UnsupportedOperationException{
      throw new UnsupportedOperationException
	("MoreGeneralIterator does not allow remove");
    }
  }

  class StackFrameMG{
    StackFrameMG(int s, DiscTree.DiscNode n){
      this.s = s;
      this.n = n;
    }
    int s;
    DiscTree.DiscNode n;
  }

  class StackFrameLG{
    StackFrameLG(int s, EndTermIterator e){
      this.s = s;
      this.e = e;
    }
    int s;
    EndTermIterator e;
  }
 


  class EndTermIterator implements Iterator{

    private boolean foundNext = false;
    private Stack back = new Stack();
    private DiscNode nextEnd;

  
    EndTermIterator(int k, DiscNode n){
      back = new Stack();
      back.push(new StackFrameET(0, k, (DiscInteriorNode) n));
    }

    public boolean hasNext(){
      if(foundNext)
	return true;
      else if (back.empty())
	return false;
      else{
	StackFrameET top = (StackFrameET) back.peek();
	while(top.i < top.n.child.length && top.n.child[top.i]==null){
	  top.i++;
	}
	if(top.i == top.n.child.length){
	  back.pop();
	  return hasNext();
	}
	else {
	  int k = 0;
	  if(top.i==0)
	    k = top.k-1;
	  else
	    k = symbolTable.arityTableInt(top.i) + top.k - 1;
	  if(k==0){
	    foundNext = true;
	    nextEnd = (DiscNode) top.n.child[top.i];
	    top.i++;
	    return true;
	  }
	  else {
	    back.push(new StackFrameET
	      (0, k, (DiscInteriorNode) top.n.child[top.i]));
	    top.i++;
	    return hasNext();
	  }
	}
      }
    }

    public Object next() throws NoSuchElementException{
      if(!hasNext())
	throw new NoSuchElementException();
      else {
	foundNext = false; //'consume' the next found endterm node
	return nextEnd;
      }
    }

    public void remove() throws UnsupportedOperationException{
      throw new UnsupportedOperationException();
    }
  }

  class StackFrameET{
    int i;
    int k;
    DiscInteriorNode n;
    StackFrameET(int i, int k, DiscInteriorNode n){
      this.i = i;
      this.k = k;
      this.n = n;
    }
  }

  class NoSuchDiscNodeException extends Exception{}

} // DiscTree



