/**
  Copyright 2002 Bruce Spencer 
*/

import java.util.*;
import java.io.*;
import org.jdom.*; 
import org.jdom.input.*; 
import org.jdom.output.*;


/**
   RuleMLFileParser is constucted given a file name to parse.  It
   creates a Vector of DefiniteClauses contained in that file.
   As it does so, it populates a SymbolTable with the symbols it
   finds.

   The Definite Clauses that are created are stored in a private
   Vector and available via the iterator, or getIputClause.

  @version $Revision: 1.1 $*/
public class RuleMLFileParser {

  public static final String IND = "ind";
  public static final String VAR = "var";
  public static final String _OPR = "_opr";
  public static final String REL = "rel";
  public static final String IMP = "imp";
  public static final String FACT = "fact";
  public static final String _HEAD = "_head";
  public static final String _BODY = "_body";
  public static final String ATOM = "atom";
  public static final String AND = "and";
  public static final String RULEBASE = "rulebase";

  /**
     Creates a new <code>RuleMLFileParser</code> instance into which 
     definite clauses can be placed.

     @param symbolTable - a <code>SymbolTable</code> value
  */
  public RuleMLFileParser(SymbolTable symbolTable) {
    inputClauses = new Vector();
    this.symbolTable = symbolTable;
  }

  /**
     The variable <code>inputClauses</code> is a Vector of
     DefiniteClause's.  A clauses parsed from the file is then
     placed into the vector into the next availabile position, as
     returned by <code>inputClauseSize()</code>.  
  */
  private Vector inputClauses; 


  /**
     The variable <code>symbolTable</code> refers to the
     SymbolTable that was given at construction time and that is
     receiving symbols as they are being parsed.  */
  private SymbolTable symbolTable;

  /**
     <code>inputClausesSize</code> returns the current size of
     the inputClauses Vector, which is taken as the next available
     position for inserting a new clause.  This is correct
     because Vectors index from zero.

     @return an <code>int</code> value representing the size*/
  public int inputClausesSize(){
    return inputClauses.size();
  }
  
  /**
     <code>getInputClause</code> returns the DefiniteClause at
     position i in clauses.  It also does the cast to
     DefiniteClause from the Vector.

     @param i - an <code>int</code> value
     @return a <code>DefiniteClause</code> value */
  public DefiniteClause getInputClause(int i){
    return (DefiniteClause) inputClauses.get(i);
  }

  /**
     <code>iterator</code> returns the iterator through the
     clauses that were read.

     @return an <code>Iterator</code> value, giving all
     DefiniteClauses that were read.  */
  public Iterator iterator(){
    return inputClauses.iterator();
  }

  /**
     <code>parseDefiniteClause</code> parses out a definite clause but
     from a String instead of from the current token stream. 
     At this moment the query comes in as a Prolog-like definite clause.
     @param clauseString - a <code>String</code> value
     @return a <code>DefiniteClause</code> value
     @exception ParseException if an error occurs */
  public DefiniteClause parseQueryClause(String clauseString) 
    throws ParseException {
    DCFileParser dcfp = new DCFileParser(symbolTable);
    return dcfp.parseQueryClause(clauseString);
  }

  private boolean compare(String s1, String s2) {
    return s1==null?s2==null:s1.equalsIgnoreCase(s2);
  }

  /**
     <code>parseRuleMLFile</code> performs a parse of a definite
     clause file and places the symbols it encounters into the
     symbolTable.

     @param fileName - a <code>String</code> value for the name
     of the file in ".ruleml" format to parse. */
  public void parseRuleMLFile(String fileName) throws JDOMException{
    // Build w/ SAX and Xerces, no validation
    SAXBuilder b = new SAXBuilder();
    // Create the document
    Document doc = b.build(new File(fileName));
    
    //      XMLOutputter outputter = new XMLOutputter("", true);
    //      try{
    //        outputter.output(doc, System.out);
    //      }catch(IOException e){e.printStackTrace();};
    Element root = doc.getRootElement();
    importDocumentRoot(root);
  }

  private void dump(Element elem){
    System.out.println("dumping " + elem);
    System.out.print("children are :");
    List list = elem.getChildren();
    Iterator li = list.iterator();
    while(li.hasNext()){
      Element o = (Element) li.next();
      System.out.println("Child " + o);
    }
  }

  private void importDocumentRoot(Element elem){
//      List list = elem.getChildren(RULEBASE);
//      for(Iterator it = list.iterator(); it.hasNext();){
//        Element next = (Element) it.next();
//        importRulebase(next);
//      }
    importRulebase(elem);
  }

  private void importRulebase(Element elem){
    List list = elem.getChildren();
    for(Iterator it = list.iterator(); it.hasNext();){
      Element next = (Element) it.next();
      if(next instanceof Element){
	if(compare(next.getName(),IMP)){
	  importImp(next);
	}
	else if (compare(next.getName(),FACT)){
	  importFact(next);
	}
      }
      else{
	//ignore elements that are not IMP or FACT
      }
    }
  }

  private void importImp(Element elem){
    //      System.out.println("Dumping imp" );
    //      dump(elem);
    currentClause = new Vector();
    vTable = new Hashtable();
    varCount = 0;
    atomCount = 0;
    Element headElement = elem.getChild(_HEAD);
    importHead(headElement);
    Element bodyElement = elem.getChild(_BODY);
    importBody(bodyElement);
    inputClauses.add(new DefiniteClause(currentClause, varCount, atomCount, symbolTable));
    vTable = null;
    currentClause = null;
  }

  private void importBody(Element elem){
    //      System.out.println("Dumping body");
    //      dump(elem);
    Element atomElement = elem.getChild(ATOM);
    if(atomElement != null){
      importAtom(atomElement);
    }
    Element conjElement = elem.getChild(AND);
    if(conjElement != null){
      List list = conjElement.getChildren(ATOM);
      for(Iterator it = list.iterator(); it.hasNext();){
	Element next = (Element) it.next();
	//  	System.out.println("Dumping what should be an atom?");
	//  	dump(next);
	importAtom(next);
      }
    }
  }

  private void importFact(Element elem){
    currentClause = new Vector();
    vTable = new Hashtable();
    varCount = 0;
    atomCount = 0;
    Element headElement = elem.getChild(_HEAD);
    importHead(headElement);
    inputClauses.add(new DefiniteClause(currentClause, varCount, atomCount, symbolTable));
    vTable = null;
    currentClause = null;
  }

  private void importHead(Element elem){
    Element atomElement = elem.getChild(ATOM);
    importAtom(atomElement);
  }

  private void importAtom(Element elem){
    int[] entry = {0, 0};
    int start = currentClause.size();
    currentClause.addElement(entry);
    Element oprElement = elem.getChild(_OPR);
    String symbol = importOpr(oprElement);
    int arity = importArgs(elem);
    int length = currentClause.size() - start;
    entry[0] = symbolTable.internSymbol(symbol, arity);
    entry[1] = length;
  }
  
  private String importOpr(Element elem){
    return importRel(elem);
  }

  private String importRel(Element elem){
    Element relElement = elem.getChild(REL);
    return relElement.getText();
  }

  private int importArgs(Element elem){
    List children = elem.getChildren();
    int nArgs = children.size() - 1;
    for(Iterator it = children.iterator(); it.hasNext();){
      Element next = (Element) it.next();
      if(next instanceof Element){
	if(compare(next.getName(), IND)){
	  importInd(next);
	}
	else if (compare(next.getName(), VAR)){
	  importVar(next);
	}
	else {
	  //do something drastic
	}
      }
    }
    return nArgs;
  }

  private void importInd(Element elem){
    String symbol = elem.getText();
    int[] entry = {0, 1};
    entry[0] = symbolTable.internSymbol(symbol,0);
    currentClause.addElement(entry);
  }

  private void importVar(Element elem){
    String varName = elem.getText();
    int i = internVar(varName);
    int[] entry = {-i, 1};
    currentClause.addElement(entry);
  }

  private Hashtable vTable;
  private int varCount;
  private int atomCount;
  private Vector currentClause;

  /** 
      This is similar to interSymbol from @see SymbolTable, but
      the hash table vhTable of variables is created anew for
      each clause, and there is no need to have a table of print
      names for variables since we do not intend to print them
      with their original names.  Thus internVar is a quick way
      to count the number of distinct vars we have seen so far in
      the clause, and to identify the ones in common. The count
      starts at 1.  New variables are referenced (named)
      according to the count that was current when the varible
      was first encountered.  Variables are put into the clause
      as negative numbers, which we get by making the additive
      complement of the reference.  Thus the first variable found
      is called -1, the second -2, etc.

      @param symbol the newly read token identifying a variable,
      which is decided by its upper case first letter
      @return an integer identifying this variable's position in
      the list of variables seen so far in this clause.  This
      list is chronologically ordered from 1.  */
  private int internVar(String symbol){
    Object varPos = vTable.get(symbol);
    if(varPos==null){ //This is a new variable.
      varCount++;
      vTable.put(symbol, new Integer(varCount));
      return varCount;
    }
    else
      return ((Integer)varPos).intValue();
  }



  public  static void main(String args[]){
    try{ 
      SymbolTable st = new SymbolTable();
      RuleMLFileParser fp = new RuleMLFileParser(st);
      fp.parseRuleMLFile("default.xml");
    
      System.out.println("SymbolTable");
      for(int i = 0; i < st.size(); i++){
	System.out.println(i + "=>" + st.printName(i));
      }
      System.out.println("Reconstruction:");
      for(int i = 0; i < fp.inputClauses.size(); i++){
	boolean debugging = false;
	DefiniteClause clause = (DefiniteClause) fp.inputClauses.get(i);
	if(debugging){
	  for(int j = 0; j < clause.symbols.length; j++){
	    if(clause.symbols[j][0] < 0){
	      System.out.println("V" + (-1-clause.symbols[j][0]));
	    
	    }
	    else
	      System.out.println(fp.symbolTable.printName(clause.symbols[j][0]) + 
				 " " + clause.symbols[j][1]);
	  }
	  System.out.println("End of clause " + i);
	}
	System.out.println(clause.toString());
      }
      
    } catch (JDOMException e){
      e.printStackTrace();
    }

  }

}




